/************************************************************************************
* This program is licensed under the BSD two-clause license, a copy of which
* is included below.
*
* Copyright (c) 2012, Claudio D. Lapilli
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* x-Redistributions of source code must retain the above copyright notice, this list
*   of conditions and the following disclaimer.
* x-Redistributions in binary form must reproduce the above copyright notice, this
*   list of conditions and the following disclaimer in the documentation and/or other
*   materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
**************************************************************************************/






#ifndef CLEANFS_H
#define CLEANFS_H


// LIBRARY CONFIGURATION OPTIONS, DEFINE IF YOU WANT TO IMPLEMENT THIS


/*!
 * \brief Make characters : * | < > ? " invalid, use \ and / as separators.
 */
#define FSCONFIG_WINDOWS_COMPAT

/*!
 * \brief Internal: Number of open files in an allocation unit
 */
#define FS_FILELIST_BLOCK  16

/*!
 * \brief File Name of the Volume's free space bitmap
 */
#define FSSTRING_FREECLUSTERS ".{~~FreeSpace~~}"
/*!
 * \brief Magic number to detect valid volumes.
 */
#define FSBOOTRECORD_MAGIC  0x53464c43      // THE STRING "CLFS"

/*!
 * \brief Type for 64-bit signed numbers.
 */
typedef long long clfs_64bit;
/*!
 * \brief Type for 32-bit signed numbers.
 */
typedef int clfs_32bit;
/*!
 * \brief Type for 16-bit signed numbers.
 */
typedef short int clfs_16bit;
/*!
 * \brief Type for 64-bit unsigned numbers.
 */
typedef unsigned long long clfs_u64bit;
/*!
 * \brief Type for 32-bit unsigned numbers.
 */
typedef unsigned int clfs_u32bit;
/*!
 * \brief Type for 16-bit unsigned numbers.
 */
typedef unsigned short int clfs_u16bit;
/*!
 * \brief Type for 8-bit characters.
 */
typedef unsigned char clfs_char;
/*!
 * \brief Type for 8-bit unsigned numbers.
 */
typedef unsigned char clfs_byte;


// FUNCTIONS THAT MUST BE SUPPLIED BY THE DISK DRIVER
/*!
 * \brief Type for driver-defined read function.
 */
typedef clfs_64bit (*driver_read_t)(clfs_64bit sector,void *buffer,clfs_64bit nsectors);
/*!
 * \brief Type for driver-defined write function.
 */
typedef clfs_64bit (*driver_write_t)(clfs_64bit sector,void *buffer,clfs_64bit nsectors);


/*!
 * \brief Bytes per directory record.
 */
#define DIRENTRY_SIZE 64    // 64 BYTES PER DIRECTORY ENTRY
/*!
 * \brief Number of bytes used for Name on a FILERECORD structure.
 */
#define DIRENTRY_NAME_LENGTH    16    // WHATEVER'S LEFT USE IT FOR THE NAME
/*!
 * \brief Number of fragments in a FRAGRECORD structure
 */
#define DIRENTRY_NUM_FRAGMENTS  5
/*!
 * \brief Limit number that separates primary records from secondary records. Entries with EntryType < PRIMARY_ENTRY_LIMIT are primary records, otherwise secondary.
 */
#define PRIMARY_ENTRY_LIMIT     256
/*!
 * \brief Number of bytes used for Name on a NAMERECORD structure.
 */
#define NAMERECORD_LENGTH       60
/*!
 * \brief Number of bytes used for Volume Label on a BOOTRECORD structure.
 */
#define BOOTRECORD_NAME_LENGTH  36

// DIRECTORY ENTRY TYPES
/*!
 * \brief Enum that defines EntryType values.
 */
enum {
    DE_EMPTY=0,             // EMPTY (UNUSED) DIRECTORY ENTRY
    DE_BOOTRECORD,
    DE_FILE,
    DE_DELETED,
    // ADD MORE PRIMARY ENTRIES HERE...

    DE_SECEMPTY=PRIMARY_ENTRY_LIMIT,
    DE_NAME,
    DE_FRAGMENT,
    //.. ADD MORE INTERNAL FILE ENTRY TYPES HERE
};

// BOOT RECORD DIRECTORY ENTRY, COULD BE FIRST ENTRY IN ROOT DIRECTORY (FIRST SECTOR OF DISK OR LAST SECTOR OF DISK)
/*!
 * \brief Boot record per specifications.
 */
struct  FS_BOOTRECORD {
    clfs_32bit EntryType;           // ALWAYS DE_BOOTRECORD
    clfs_32bit Magic;               // ALWAYS 0x = "CLFS"
    clfs_32bit BytesPerCluster;     // CLUSTER SIZE IN BYTES
    clfs_64bit TotalClusters;       // VOLUME TOTAL SIZE
    clfs_64bit RootDirCluster;      // NORMALLY 0 UNLESS ROOT DIR WAS RELOCATED DUE TO DISK FAILURE, NOTE THAT BOOT RECORD MIGHT BE INSIDE ROOT DIRECTORY
    clfs_char VolName[BOOTRECORD_NAME_LENGTH];          // IF VOLUME NAME IS MORE THAN 36 CHARACTERS, ONE OR MORE NAME_ENTRIES MAY FOLLOW
};
/*!
 * \brief File record per specifications.
 */
struct FS_FILERECORD {
    clfs_32bit EntryType;           // ALWAYS DE_FILE
    clfs_32bit Attributes;          // SEE FILE ATTRIBUTES
    clfs_64bit FileSize;            // SIZE IN BYTES
    clfs_64bit StartCluster;        // FIRST FRAGMENT OF THE FILE, ONE OR MORE DE_FRAGMENT MIGHT FOLLOW FOR FRAGMENTED FILES
    clfs_32bit NumClusters;         // SIZE OF FIRST FRAGMENT
    clfs_32bit CreateDate;          // CREATION TIME AND DATE IN DOS-FORMAT
    clfs_32bit ModifiedDate;        // LAST MODIFICATION TIME AND DATE
    clfs_32bit ExtendedDates;       // HIGH PRECISION CREATION/MODIFIED DATE
    clfs_32bit NameCRC;             // CRC32 OF THE NAME
    clfs_32bit FoldedCRC;           // CRC32 OF THE CASE-FOLDED NAME
    clfs_char  Name[DIRENTRY_NAME_LENGTH];            // FILE NAME, IF MORE THAN 27 CHARACTERS, ONE OR MORE DE_NAME ENTRIES WILL FOLLOW.
};


/*!
 * \brief Fragments record per specifications.
 */
struct FS_FRAGRECORD {
    clfs_32bit EntryType;           // ALWAYS DE_FRAGMENT
    struct {
            clfs_64bit StartCluster;
            clfs_32bit NumClusters;                         // SET TO ZERO INDICATES UNUSED ENTRY
    } Fragments[DIRENTRY_NUM_FRAGMENTS];                     // UP TO 5 FRAGMENTS PER DIRECTORY ENTRY, ONE OR MORE DE_FRAGMENT ENTRIES MAY FOLLOW
};
/*!
 * \brief Name record per specifications.
 */
struct FS_NAMERECORD {
    clfs_32bit EntryType;                           // ALWAYS DE_NAME
    clfs_char  Name[NAMERECORD_LENGTH];            // PART OF A NAME ONE OR MORE DE_NAME ENTRIES WILL FOLLOW. NAME STRING IS NULL-TERMINATED.
};
/*!
 * \brief Arbitrary Metadata record per specifications.
 */
struct FS_METARECORD {
    clfs_32bit EntryType;           // ANY SECONDARY TYPE
    clfs_byte  Metadata[60];        // ARBITRARY METADATA
};


struct FS_FILE;
struct FS_FILELIST;
struct FS_VOLUME;
struct FS_FRAGMENT;

/*!
 * \brief Data structure to encapsulate disk information.
 */
struct FS_DISK {
    /*!
     * \brief Function supplied by the driver to read sectors.
     */
    driver_read_t read;
    /*!
     * \brief Function supplied by the driver to write sectors.
     */
    driver_write_t write;
    /*!
     * \brief Total number of sectors in the disk.
     */
    clfs_64bit TotalSectors;
    /*!
     * \brief Number of the first sector in the disk (LBA sector at beginning of partition in a hard disk).
     */
    clfs_64bit FirstSector;
    /*!
     * \brief Size in bytes of a sector. Normally 512 bytes on legacy hard drives, 4096 bytes on newer drives.
     */
    clfs_32bit SectorSize;
};

/*!
 * \brief Linked list of FS_FILE pointers to open files.
 */
struct FS_FILELIST {
    struct FS_FILE *File[FS_FILELIST_BLOCK];
    clfs_32bit UsedCount;
    struct FS_FILELIST *Next;
};


/*!
 * \brief Internal buffer management constant.
 */
#define FSBUFFER_HASDATA    1
/*!
 * \brief Internal buffer management constant.
 */
#define FSBUFFER_DIRTY      2

/*!
 * \brief Internal buffer data structure.
 */
struct FS_BUFFER {
    clfs_byte *Data;
    clfs_32bit Used;
    clfs_64bit Offset;
    };



/*!
 * \brief Linked list of directory records.
 */
struct FS_RECORDLIST {
    union {
        struct FS_FILERECORD Filerec;
        struct FS_NAMERECORD Namerec;
        struct FS_FRAGRECORD Fragrec;
        struct FS_METARECORD Metarec;
    };
    struct FS_RECORDLIST *Next;
};

/*!
 * \brief End User opaque data structure that encapsulates a file.
 */
struct FS_FILE {
    /*!
     * The FILERECORD of this file
     */
    struct FS_FILERECORD F;

    /*!
     * Access mode. Any combination of the FSMODE_XXX constants
     */
    clfs_32bit   Mode;
    /*!
    * Current file offset pointer
    */
    clfs_64bit CurrentOffset;
    /*!
     * Volume that contains this file
     */
    struct FS_VOLUME *Volume;
    /*!
     * Pointer to parent directory
     */
    struct FS_FILE *Dir;
    /*!
     * Offset of file entry within its parent directory
     */
    clfs_64bit DirEntryOffset;
    /*!
     * Total Number of entries within its parent directory
     */
    clfs_32bit DirEntryCount;
    /*!
     * Offset of directory entry within itself (for directories only)
     */
    clfs_64bit DirEntryOffset2;
    /*!
     * Total Number of entries within itself (for directories only)
     */
    clfs_32bit DirEntryCount2;
    /*!
     * Compact version of name
     */
    clfs_char *Name;
    /*!
     * **Internal**: Read/Write buffers
     */
    struct FS_BUFFER Buffer;

    /*!
     * **Internal**: Linked list of fragments of this file
     */
    struct FS_RECORDLIST *Chain;
    /*!
     * **Internal**: Linked list of metadata records attached to this file
     */
    struct FS_RECORDLIST *Metadata;     // ALL OTHER METADATA RECORDS GO HERE

};


/*!
 * \brief End User opaque data structure that encapsulates a disk volume.
 */
struct FS_VOLUME {
    /*!
     * \brief Information about the physical disk, including read/write functions.
     */
    struct FS_DISK Disk;
    /*!
     * \brief Size in bytes of a cluster.
     */
    clfs_64bit ClusterSize;
    /*!
     * \brief Total Number of clusters in the volume.
     */
    clfs_64bit NumClusters;
    /*!
     * \brief Number of sectors per cluster.
     */
    clfs_32bit SectorsPerCluster;
    /*!
     * \brief Number of bits n such that ClusterSize=2^n, or ClusterSize=1<<n;.
     */
    clfs_32bit ClusterBits;
    /*!
     * \brief Flags indicating various states of the volumn.
     */
    clfs_32bit Flags;
    /*!
     * \brief Name of the volume as a UTF-8 string.
     */
    clfs_char *VolName;
    /*!
     * \brief The root directory of the volume.
     */
    struct FS_FILE RootDir;
    /*!
     * \brief Pointer to the current work directory.
     */
    struct FS_FILE *CurrentDir;
    /*!
     * \brief Pointer to the free space bitmap file.
     */
    struct FS_FILE *FreeClusters;
    /*!
     * \brief Hint indicating which cluster is the next likely empty cluster in the volume.
     */
    clfs_64bit NextFreeClusterHint;
    /*!
     * \brief List of all files open.
     */
    struct FS_FILELIST *OpenFiles;

};


#define FS_FALSE    0
#define FS_TRUE     (!FS_FALSE)


// ERRORS RETURNED BY USER-LEVEL FUNCTIONS
/*!
 * \brief Error code: Returned by all functions when there are no errors.
 */
#define FS_OK		  0

/*!
 * \brief Error code: Indicates end-of-file was reached
 */
#define FS_EOF	     -1			// END OF FILE
/*!
 * \brief Error code: Indicates an invalid file name was given
 */
#define FS_BADNAME   -2			// INVALID FILE NAME
/*!
 * \brief Error code: Returned when the volume contains invalid or corrupted data structures
 */
#define FS_INVALIDDATA -3			// INVALID DATA
/*!
 * \brief Error code: File not found.
 */
#define FS_NOTFOUND  -4			// FILE NOT FOUND
/*!
 * \brief Error code: Write operation failed or not permitted.
 */
#define FS_CANTWRITE -5			// WRITE FAILED
/*!
 * \brief Error code: No disk is currently inserted.
 */
#define FS_NODISK    -6			// NO DISK INSERTED
/*!
 * \brief Error code: User has changed the card without unmounting the volumes.
 */
#define FS_CHANGED   -7 		// CARD HAS CHANGED
/*!
 * \brief Error code: Max. number of open files has been reached.
 */
#define FS_MAXFILES  -8			// MAXIMUM NUMBER OF FILES OPEN WAS EXCEEDED
/*!
 * \brief Error code: The name given correspond to an open directory.
 */
#define FS_OPENDIR   -9			// NAME IS AN ALREADY OPEN DIRECTORY
/*!
 * \brief Error code: The name given corresponds to an open file.
 */
#define FS_OPENFILE  -9			// NAME IS AN ALREADY OPEN FILE
/*!
 * \brief Error code: Disk is full.
 */
#define FS_DISKFULL  -10		// DISK IS FULL
/*!
 * \brief Error code: File already exists.
 */
#define FS_EXIST     -11		// FILE ALREADY EXISTS
/*!
 * \brief Error code: Insufficient memory.
 */
#define FS_NOMEM     -12		// INSUFFICIENT MEMORY TO COMPLETE OPERATION
/*!
 * \brief Error code: Cannot read disk.
 */
#define FS_CANTREAD  -13		// DISK READ ERROR
/*!
 * \brief Error code: Indicates a hardware error or an unknown error.
 */
#define FS_ERROR      -14		    // UNKNOWN ERROR (OR FUNCTION DOESN'T CARE)
/*!
 * \brief Error code: File/directory is in use.
 */
#define FS_USED       -15			// FILE/DIRECTORY IS BEING USED




// OPEN MODES
/*!
 * \brief Read only mode
 *
 * Write operations won't be allowed on this file.
 */
#define FSMODE_READ      0    // READ PERMISSION (DEFAULT)
/*!
 * \brief Write mode
 *
 * Write operations truncate the file after the last written byte.
 */
#define FSMODE_WRITE     2    // WRITE PERMISSION
/*!
 * \brief Append mode
 *
 * Write operations occur only at the end of the file.
 */
#define FSMODE_APPEND    4    // ONLY ALLOW WRITING AFTER EOF
/*!
 * \brief Modify mode
 *
 * Write operations occur anywhere in the file. File preserves its
 * original size unless written beyond end-of-file.
 */
#define FSMODE_MODIFY    8    // WRITE TO ANY POSITION WITHIN FILE
/*!
 * \brief No-growth mode
 *
 * Write operations that require to increase the file size won't
 * be allowed.
 */
#define FSMODE_NOGROW   16    // DON'T ALLOW FILE TO GROW
/*!
 * \brief Don't create file
 *
 * This flag is for the FSOpen function only. If write permission was
 * requested and the file doesn't exist, don't create a new one.
 */
#define FSMODE_NOCREATE	32    // DON'T CREATE IF FILE DOESN'T EXIST

/*!
 * \brief INTERNAL: do not use
 *
 * This flag is internally used and should not be set. Force a file or
 * directory to fully update its directory entry.
 */
#define FSMODE_DIRTYDIR	1024



// FILE ATTRIBUTES
/*!
 * \brief Directory file attribute (set on directories only)
 */
#define FSATTR_DIR    1
/*!
 * \brief Hidden file attribute
 */
#define FSATTR_HIDDEN 2
/*!
 * \brief Immutable file attribute
 */
#define FSATTR_IMMUTABLE 4
/*!
 * \brief File is case insensitive
 */
#define FSATTR_CASEINSENS 8
/*!
 * \brief Directory contents are case insensitive by default
 */
#define FSATTR_DIRCASEINSENS 16




// ADD MORE ATTRIBUTES HERE...





// NAME ANALYSIS CONSTANTS

// RETURN TYPE OF PATH
// 1 == Name is empty
// 2 == Name include path
// 4 == Path is absolute
// 8 == File ends in slash
// BITS 8-31 = DEPTH OF PATH (0=CURRENT DIR OR ROOT)
// NEGATIVE RESULT ==> ERROR CODE (NORMALLY FS_BADNAME)

/*!
 * \brief Internal: Given text has an empty file name
 */
#define FSNAME_EMPTY 1
/*!
 * \brief Internal: Given text includes a path
 */
#define FSNAME_HASPATH 2
/*!
 * \brief Internal: Given text is an absolute path
 */
#define FSNAME_ISABSOLUTE 4
/*!
 * \brief Internal: Given text ends with a slash character
 */
#define FSNAME_ENDSLASH 8






// DEFINE THESE CONSTANTS HERE TO AVOID DEPENDENCY ON stdio.h
#ifndef SEEK_SET
/*!
 * \brief Start a seek from the first byte of the file.
 */
#define SEEK_SET 0
#endif
#ifndef SEEK_CUR
/*!
 * \brief Start a seek from the current position within the file.
 */
#define SEEK_CUR 1
#endif
#ifndef SEEK_END
/*!
 * \brief Start a seek from the last byte of the file.
 */
#define SEEK_END 2
#endif

// OS-DEPENDENT FUNCTIONS. REIMPLEMENT FOR EACH OS.
/*!
 * \brief Get Current Time and Date
 *
 * This function is OS-dependant. It must be reimplemented for each target.
 *
 * \return 64-bit number that contains the time elapsed since 1/1/1970 (Unix time) expressed in 1/65536th of second.
 * In systems without a high resolution timer, it can be (n<<16) where n=Number of seconds since 1/1/1970, which is
 * the standard output of the POSIX time() function. So the implementation on a POSIX system can simply be (time()<<16).
 *
 * \sa FSMalloc FSFree
 */

extern clfs_64bit FSTimeDateNow();

/*!
 * \brief Allocates a block of memory. Similar to standard C malloc().
 *
 * This function is OS-dependant. It must be reimplemented for each target.
 *
 * \param size The requested block size in bytes (64-bit number).
 * \return A pointer to the newly allocated block, or NULL if no block was allocated. On standard systems it can simply call malloc().
 *
 * \sa FSTimeDateNow FSFree
 */
extern void *FSMalloc(clfs_64bit size);

/*!
 * \brief Releases a block of memory. Similar to standard C free().
 *
 * This function is OS-dependant. It must be reimplemented for each target.
 *
 * \param ptr A pointer to a memory block previously allocated with FSMalloc().
 *
 * \sa FSTimeDateNow FSMalloc
 */

extern void FSFree(void *ptr);




// LOW-LEVEL CLUSTER OPERATIONS

/*!
 * \brief Read a group of clusters into memory.
 *
 * Reads a contiguous group of clusters directly into memory.
 *
 * \param cluster Cluster number of the first cluster to read.
 * \param buffer  Pointer to a previously allocated buffer where the clusters will be stored. The buffer
 * must have enough space to store nclusters.
 * \param nclusters Number of clusters to transfer
 * \param fs Pointer to the volume to read clusters from.
 *
 * \return The number of bytes effectively read from disk. If no errors occurred, it will match nclusters*ClusterSize.
 *
 * \sa FSWriteCluster
 */
extern clfs_64bit FSReadCluster(clfs_64bit cluster,clfs_byte *buffer,clfs_64bit nclusters,struct FS_VOLUME *fs);
/*!
 * \brief Write a group of clusters into memory.
 *
 * Writes a contiguous group of clusters directly into disk.
 *
 * \param cluster Cluster number of the first cluster to write.
 * \param buffer  Pointer to a previously allocated buffer containing the data to be written.
 * \param nclusters Number of clusters to transfer.
 * \param fs Pointer to the volume to write clusters to.
 *
 * \return The number of bytes effectively written to disk. If no errors occurred, it will match nclusters*ClusterSize.
 *
 * \sa FSReadCluster
 */
extern clfs_64bit FSWriteCluster(clfs_64bit cluster,clfs_byte *buffer,clfs_64bit nclusters,struct FS_VOLUME *fs);


/*!
 * \brief Error messages text strings.
 *
 */
// ERROR MESSAGE STRINGS
extern const char * const FSErrorMsg[];




// BUFFER MANAGEMENT FUNCTIONS

/*!
 * \brief Initialize buffers for the given file.
 *
 * Allocates memory and initializes the read/write buffers for the given file/directory.
 * Mainly for internal use.
 *
 * \param file Pointer to a FS_FILE structure containing the file information.
 *
 * \return 0 (FS_OK) if no errors occurred, an error code otherwise.
 *
 * \sa FSBufferFlush FSBufferFill FSBufferMap FSBufferMarkDirty
 */
extern int FSInitFileBuffers(struct FS_FILE *file);


/*!
 * \brief Flush the buffers for the given file.
 *
 * If the buffers contain any data that has not been written to disk, it writes
 * the data.

 * \param file Pointer to a FS_FILE structure containing the file information.
 *
 * \return 0 (FS_OK) if no errors occurred, an error code otherwise. If the buffer wasn't dirty, it returns FS_OK.
 *
 * \sa FSInitFileBuffers FSBufferFill FSBufferMap FSBufferMarkDirty
 */
extern int FSBufferFlush(struct FS_FILE *file);

/*!
 * \brief Fill the buffers with data from the given file.
 *
 * If the buffers contain any data that has not been written to disk, it writes
 * the data first. The buffer is filled with the cluster that contains NewOffset.
 *
 * \param file Pointer to a FS_FILE structure containing the file information.
 * \param NewOffset Offset to fill the buffer from. If offset is not a multiple of the cluster size, it will be
 * rounded down to the nearest cluster boundary.
 *
 * \return 0 (FS_OK) if no errors occurred, an error code otherwise. When it returns FS_OK, the buffer is
 * guaranteed to contain data at NewOffset, but not necessarily at the beginning of the buffer (due to cluster
 * boundary rounding).
 *
 * \sa FSInitFileBuffers FSBufferFlush FSBufferMap FSBufferMarkDirty
 */

extern int FSBufferFill(struct FS_FILE *file,clfs_64bit NewOffset);

/*!
 * \brief Mark the contents of a buffer as dirty.
 *
 * Marking the buffer dirty signals that the buffer contains information ready to be written to disk.
 * This function is only used internally, when the application writes data manually into the buffer.
 *
 * \param file Pointer to a FS_FILE structure containing the file information.
 *
 * \return Nothing.
 *
 * \sa FSInitFileBuffers FSBufferFill FSBufferMap FSBufferFlush
 */
extern inline void FSBufferMarkDirty(struct FS_FILE *file);

/*!
 * \brief Fill the buffers with data from the file at Offset, and return a pointer into the buffer.
 *
 * This function fills the buffer with the proper cluster, then returns a pointer that corresponds
 * to Offset within the buffer. At least one byte is guaranteed to be in the buffer at Offset.
 * This is an internal function used mainly to scan directories without copying the data to a separate location.
 * The code using this function must account for cluster boundaries and the limited buffer size.
 *
 * \param file Pointer to a FS_FILE structure containing the file information.
 * \param Offset Offset within the file to get a pointer to.
 *
 * \return a clfs_byte pointer into the internal buffers, pointing to the byte corresponding to Offset. It
 * returns NULL if there's an error.
 *
 *
 * \sa FSInitFileBuffers FSBufferFill FSBufferMarkDirty FSBufferFlush
 */
extern clfs_byte *FSBufferMap(struct FS_FILE *file,clfs_64bit Offset);

// CLUSTER ALLOCATION FUNCTIONS
extern struct FS_RECORDLIST *FSAllocClusters(clfs_64bit requestedcluster,clfs_64bit needed,struct FS_VOLUME *fs);
extern clfs_64bit FSGetLastCluster(struct FS_FILE *file);
extern clfs_64bit FSOffsetToCluster(clfs_64bit Offset,struct FS_FILE *file);
extern void FSGetFragmentPtr(clfs_64bit Offset,struct FS_FILE *file,struct FS_RECORDLIST **frag,clfs_32bit *idx);
extern void FSGetNextFragment(struct FS_RECORDLIST **frag,clfs_32bit *idx);
extern void FSAppendFragments(struct FS_FILE *file,struct FS_RECORDLIST *fragments);
// FREE SPACE MANAGEMENT FUNCTIONS
extern inline void FSAnalyzeUsage(clfs_32bit word,clfs_32bit *left,clfs_32bit *mid,clfs_32bit *midpos,clfs_32bit *right);
extern clfs_64bit FSNextFreeCluster(struct FS_VOLUME *fs);
extern int FSMarkClustersUsed(clfs_64bit Cluster,clfs_64bit NumClusters,struct FS_VOLUME *fs);
extern int FSMarkClustersFree(clfs_64bit Cluster,clfs_64bit NumClusters,struct FS_VOLUME *fs);
extern int FSRescanUsedClusters(struct FS_FILE *Dir);


// METADATA HANDLING FUNCTIONS
extern void FSAppendMetadata(struct FS_FILE *file,struct FS_RECORDLIST *meta);


// STRING FUNCTIONS
extern clfs_u32bit FSCalcNameCRC(clfs_char * data);
extern clfs_u32bit FSCalcFoldedNameCRC(clfs_char * data);
extern int FSPartialNameCompare(clfs_char *name,clfs_char *entryName,clfs_32bit *CharacterCache,clfs_char **nameptr,clfs_32bit entrychars,int CaseInsens);
extern int FSNameCompare(clfs_char *name,clfs_char *othername,int CaseInsens);
extern void FSMemset(void *dest,clfs_32bit value,clfs_32bit nbytes);
extern void FSMemcpy(void *dest,void *src,clfs_32bit nbytes);
extern int FSGetNameType(clfs_char *name);



// DIRECTORY ENTRY FUNCTIONS
extern int FSFindEntry(clfs_char *name,int CaseInsens, struct FS_FILE *record,struct FS_FILE *dir);
extern int FSGetDirEntryData(struct FS_FILE *dir,clfs_64bit Offset,struct FS_FILE *file);
extern void FSFreeEntryData(struct FS_FILE *file);
extern void FSFreeEntry(struct FS_FILE *file);
extern void FSDirEntryMoved(struct FS_FILE *dir,clfs_64bit OrgOffset,clfs_64bit EndOrgOffset, clfs_64bit NewOffset);
extern int FSAllocDirEntries(struct FS_FILE *dir,clfs_32bit numentries,clfs_64bit *DirEntryOffset);
extern clfs_32bit FSGetDirEntryRecordCount(struct FS_FILE *dir,clfs_64bit DirEntryOffset);
extern clfs_64bit FSSkipDirEntry(struct FS_FILE *dir,clfs_64bit DirEntryOffset);
extern clfs_64bit FSRSkipDirEntry(struct FS_FILE *dir,clfs_64bit DirEntryOffset);
extern int FSMoveDirEntryToEnd(struct FS_FILE *dir,clfs_64bit DirEntryOffset);
extern int FSAllocUnmovableDirEntries(struct FS_FILE *dir,clfs_32bit numentries,clfs_64bit DirEntryOffset);
extern int FSUpdateDirEntry(struct FS_FILE *file,struct FS_FILERECORD *newinfo);
extern int FSGetNextDirEntry(struct FS_FILE *Dir,struct FS_FILE **nextentry);
extern clfs_32bit FSCalcDirEntryCount(struct FS_FILE *file,int *__singlefrag);
extern int FSRenameDirEntry(struct FS_FILE *file,clfs_char *newname);



// VOLUME FUNCTIONS
extern int FSMountVolume(struct FS_DISK *Disk,struct FS_VOLUME *fs);
extern int FSUnmountVolume(struct FS_VOLUME *fs,clfs_32bit forceclose);
extern int FSMountFreeSpace(struct FS_VOLUME *fs);
extern int FSRescanFreeSpace(struct FS_VOLUME *fs);



// DIRECTORY FUNCTIONS
extern int FSOpenDirLL(struct FS_FILE *Dir,struct FS_VOLUME *fs,clfs_64bit FirstCluster);
extern int FSOpenDir(clfs_char *name,struct FS_FILE **fileptr,struct FS_VOLUME *fs);
extern int FSMkdir(clfs_char *name,struct FS_VOLUME *fs);
extern int FSRmdir(clfs_char *name,int force,struct FS_VOLUME *fs);
extern int FSChdir(clfs_char *name,struct FS_VOLUME *fs);
extern int FSGetcwd(clfs_char **bufferptr, struct FS_VOLUME *fs);


// FILE FUNCTIONS
extern int FSCreate(clfs_char *name,clfs_32bit attr,struct FS_FILE **fileptr,struct FS_VOLUME *fs);
extern int FSOpen(clfs_char *name, clfs_32bit mode, struct FS_FILE **fileptr, struct FS_VOLUME *fs);
extern int FSClose(struct FS_FILE *file);
extern clfs_64bit FSRead(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file);
extern clfs_64bit FSWrite(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file);
extern clfs_64bit FSReadLL(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file);
extern clfs_64bit FSWriteLL(clfs_byte *buffer,clfs_64bit nbytes,struct FS_FILE *file);
extern int FSSeek(struct FS_FILE *file,clfs_64bit Offset,clfs_32bit position);
extern int FSGrowFile(clfs_64bit filesize,struct FS_FILE *file);
extern int FSFindFile(clfs_char *name,struct FS_FILE *entry, struct FS_VOLUME *fs,clfs_32bit dirsvalid);
extern struct FS_FILE *FSFreeFile(struct FS_FILE *file);
extern int FSFileIsReferenced(struct FS_FILE *file,struct FS_VOLUME *fs);
extern int FSFileIsOpen(struct FS_FILE *file,struct FS_VOLUME *fs);
extern int FSAddToOpenFileList(struct FS_FILE *entry);
extern int FSRemoveFromOpenFileList(struct FS_FILE *entry);
extern int FSRename(clfs_char *oldfile,clfs_char *newname,struct FS_VOLUME *fs);
extern int FSDelete(clfs_char *name,struct FS_VOLUME *fs);
extern void FSDeleteOnClose(struct FS_FILE *file);


// INTERNAL FUNCTIONS
extern clfs_64bit FSOffsetToCluster(clfs_64bit Offset,struct FS_FILE *file);
extern clfs_64bit FSNextFreeCluster(struct FS_VOLUME *fs);
extern clfs_u32bit FSCalcNameCRC(clfs_char * data);
extern clfs_u32bit FSCalcFoldedNameCRC(clfs_char * data);




// USER-LEVEL INLINE FUNCTIONS TO RETURN FILE PROPERTIES
extern inline clfs_32bit FSFileAttributes(struct FS_FILE *file) { return file->F.Attributes; }
extern inline clfs_32bit FSFileType(struct FS_FILE *file) { return file->F.EntryType; }
extern inline clfs_64bit FSFileSize(struct FS_FILE *file) { return file->F.FileSize; }
extern inline int FSEof(struct FS_FILE *file) { return (file->CurrentOffset>=file->F.FileSize)? FS_TRUE:FS_FALSE; }



#endif // CLEANFS_H
