/************************************************************************************
* This program is licensed under the BSD two-clause license, a copy of which
* is included below.
*
* Copyright (c) 2012, Claudio D. Lapilli
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* x-Redistributions of source code must retain the above copyright notice, this list
*   of conditions and the following disclaimer.
* x-Redistributions in binary form must reproduce the above copyright notice, this
*   list of conditions and the following disclaimer in the documentation and/or other
*   materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
**************************************************************************************/




#include <stdlib.h>
#include <time.h>

#include "cleanfs.h"

#define FSDEBUG

#include "implement.h"

// ONLY EXTERNAL DEPENDENCIES ARE:
// malloc() / free()
// time()

// THIS IS THE ONLY FUNCTION THAT HAS A DEPENDENCY ON time()
// IT IS ENCAPSULATED FOR EASY REPLACEMENT
// THIS FUNCTION MUST ALWAYS RETURN NUMBER OF SECONDS SINCE 1/1/1970

clfs_64bit FSTimeDateNow()
{
    time_t t=time(NULL);
    return ((clfs_64bit)t)<<16;
}

// THIS IS THE ONLY FUNCTION THAT HAS A DEPENDENCY ON malloc()
// IT IS ENCAPSULATED FOR EASY REPLACEMENT

void *_FSMalloc(clfs_64bit size)
{
    return malloc((size_t) size);
}
// THIS IS THE ONLY FUNCTION THAT HAS A DEPENDENCY ON free()
// IT IS ENCAPSULATED FOR EASY REPLACEMENT
void _FSFree(void *ptr)
{
    free(ptr);
}

#ifdef FSDEBUG


#define MAX_MALLOC_BLOCKS 2048

struct {
    void *ptr;
    char *filename;
    int line;
    long long size;
} memblocks[MAX_MALLOC_BLOCKS];
int memblkcount=0,maxblkcount=0;


void *FSMallocDebug(clfs_64bit size,const char *filename,int line)
{

    memblocks[memblkcount].ptr=malloc((size_t) size);
    if(memblocks[memblkcount].ptr) {
        memblocks[memblkcount].filename=(char *)filename;
        memblocks[memblkcount].line=line;
        memblocks[memblkcount].size=size;
        int prevblk=memblkcount;
        ++memblkcount;
        memblkcount&=(MAX_MALLOC_BLOCKS-1);  // CREATE CIRCULAR BUFFER JUST IN CASE

        if(!memblkcount) dbg_msg("WARNING: Exceeded block count limit on malloc.\n");

        if(maxblkcount<memblkcount) maxblkcount=memblkcount;

                return memblocks[prevblk].ptr;
    }
    return NULL;
}


void FSFreeDebug(void *ptr,const char *filename,int line)
{
    int i;
    for(i=0;i<memblkcount;++i)
    {
        if(ptr==memblocks[i].ptr) {
            memblocks[i].ptr=NULL;  // MARK IT AS FREE
            break;
        }
    }
    if(i==memblkcount) dbg_msg("WARNING: freeing nonexistent block: %s, line %d\n",filename,line);

    // SHRINK THE LIST
    while(memblkcount) {
        --memblkcount;
        if(memblocks[memblkcount].ptr!=NULL) { ++memblkcount; break; }
    }

    free(ptr);
}

void FSMemreport()
{
    int i;
    // REPORT
    dbg_msg("Total allocated blocks=%d\n",memblkcount);
    dbg_msg("Max. allocated blocks=%d\n",maxblkcount);
    dbg_msg("\nPTR        SIZE        LINE   FILE\n");
    dbg_msg("Using %d bit pointers\n",sizeof(void *)*8);
    dbg_msg(  "------------------------------------------------------\n");
    for(i=0;i<memblkcount;++i)
    {
        dbg_msg("0x%08X %11lld %6d %s\n",(unsigned int)(memblocks[i]).ptr,memblocks[i].size,memblocks[i].line,memblocks[i].filename);
    }
    dbg_msg("End of List.\n");
}



#define FSMalloc(a) FSMallocDebug(a,__FILE__,__LINE__)
#define FSFree(a)   FSFreeDebug(a,__FILE__,__LINE__)

#else

#define FSMalloc(a) _FSMalloc(a)
#define FSFree(a) _FSFree(a)

#endif
