/************************************************************************************
* This program is licensed under the BSD two-clause license, a copy of which
* is included below.
*
* Copyright (c) 2012, Claudio D. Lapilli
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* x-Redistributions of source code must retain the above copyright notice, this list
*   of conditions and the following disclaimer.
* x-Redistributions in binary form must reproduce the above copyright notice, this
*   list of conditions and the following disclaimer in the documentation and/or other
*   materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
**************************************************************************************/




#include <QtCore/QCoreApplication>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

extern "C" {
#include "cleanfs.h"
#define FSDEBUG
#include "implement.h"      // ONLY NEEDED TO USE THE DEBUGGING FUNCTION FSMemreport()
#include "cleanfstools.h"
}










#define TEST_DISK_FILE "./TESTFILE.DAT"
#define TEST_DISK_SIZE 64LL   // in MB




// DISK DRIVER FUNCTIONS - FOR TEST PURPOSES ONLY

char *TestBuffer;

int DRV_OpenDisk()
{

    TestBuffer=(char *)malloc(TEST_DISK_SIZE*1024LL*1024LL);
    if(!TestBuffer) return false;

    // CHECK IF TEST FILE EXISTS

    int han=open(TEST_DISK_FILE,O_RDONLY);

    if(han!=-1) {
        read(han,TestBuffer,TEST_DISK_SIZE*1024LL*1024LL);
        close(han);
    }

        return true;

}

int DRV_CloseDisk()
{

    int han=open(TEST_DISK_FILE,O_RDWR | O_CREAT| O_TRUNC,S_IRWXU );

    if(han!=-1) {
        write(han,TestBuffer,TEST_DISK_SIZE*1024LL*1024LL);
        close(han);
    }

    if(TestBuffer) free(TestBuffer);
    TestBuffer=NULL;
    return true;
}

clfs_64bit DRV_Read(clfs_64bit sector,void *buffer,clfs_64bit nsectors)
{
    if(sector<0) return 0;
    if(sector+nsectors>TEST_DISK_SIZE*1024LL*2) { nsectors=TEST_DISK_SIZE*1024LL*2-sector; }
    if(nsectors<=0) return 0;
    memmove(buffer,&(TestBuffer[sector*512]),nsectors*512);
    return nsectors*512;
}

clfs_64bit DRV_Write(clfs_64bit sector,void *buffer,clfs_64bit nsectors)
{
    if(sector<0) return 0;
    if(sector+nsectors>TEST_DISK_SIZE*1024LL*2) { nsectors=TEST_DISK_SIZE*1024LL*2-sector; }
    if(nsectors<=0) return 0;
    memmove(&(TestBuffer[sector*512]),buffer,nsectors*512);
    return nsectors*512;
}



void DRV_GetDiskInfo(FS_DISK *dsk)
{
    // Clear data structure
    memset(dsk,0,sizeof(FS_DISK));

    dsk->FirstSector=0;
    dsk->SectorSize=512;
    dsk->TotalSectors=TEST_DISK_SIZE*1024LL*2;
    dsk->read=&DRV_Read;
    dsk->write=&DRV_Write;
    return;
}


// SIMPLE SHELL COMMANDS FOR TESTING PURPOSES

typedef void (*CMD_FUNC)(char *);

void sh_exit(char *args);
void sh_mount(char *args);
void sh_unmount(char *args);
void sh_cd(char *args);
void sh_mkdir(char *args);
void sh_pwd(char *args);
void sh_dir(char *args);
void sh_format(char *args);
void sh_mem(char *args);
void sh_check(char *args);
void sh_ls(char *args);
void sh_dotest(char *args);
void sh_cat(char *args);
void sh_rename(char *args);
void sh_remove(char *args);
void sh_rmdir(char *args);


typedef struct {
    char *cmd;
    CMD_FUNC func;
} CMD_ITEM;

CMD_ITEM CmdList[]={
    {"exit",&sh_exit},
    {"mount",&sh_mount},
    {"unmount",&sh_unmount},
    {"cd",&sh_cd },
    {"dir",&sh_dir},
    {"format",&sh_format},
    {"pwd",&sh_pwd},
    {"mkdir",&sh_mkdir},
    {"mem",&sh_mem},
    {"check",&sh_check},
    {"ls",&sh_ls},
    {"dotest",&sh_dotest},
    {"cat",&sh_cat},
    {"rename",&sh_rename},
    {"rm",&sh_remove},
    {"rmdir",&sh_rmdir},

    //... ADD COMMANDS HERE
    {"",NULL}
};

struct st_shell_data {
    struct FS_DISK dsk;
    struct FS_VOLUME fs;
    int is_mounted,terminate;
} shell_data;





// TESTING UNIT FOR FILE SYSTEM PERFORMANCE

int main(int argc, char *argv[])
{
    char cmdbuf[1024];
    int space,cmdidx;

    memset(&shell_data,0,sizeof(shell_data));

    printf("CleanFS Test console\n\n");
    while(!shell_data.terminate)
    {
        printf("\n#");
        gets(cmdbuf);
        space=0;
        while( (cmdbuf[space]!=' ') && (cmdbuf[space]!=0)) ++space;
        // FIND COMMAND

        for(cmdidx=0;CmdList[cmdidx].func!=NULL;++cmdidx)
        {
            if(strlen(CmdList[cmdidx].cmd)==space) {
                if(!strncmp(CmdList[cmdidx].cmd,cmdbuf,space)) {
                    // COMMAND FOUND, EXECUTE
                    CmdList[cmdidx].func(cmdbuf+space+1);
                    break;
                }
            }
        }
        if(CmdList[cmdidx].func==NULL) {
            printf("Unknown command.\n");
        }

    }
    return 0;
}


void sh_error(int error)
{
    printf("Error: ");
    switch(error)
    {
    case FS_OK:
        printf("OK.\n");
        break;
    case FS_EOF:
        printf("End of file.\n");
        break;
    case FS_BADNAME:
        printf("Bad name.\n");
        break;
    case FS_INVALIDDATA:
        printf("Volume contains invalid or corrupted data structures.\n");
        break;
    case FS_NOTFOUND:
        printf("File not found.\n");
        break;
    case FS_CANTWRITE:
        printf("Write operation failed or not permitted.\n");
        break;
    case FS_NODISK:
        printf("No disk is currently inserted.\n");
        break;
    case FS_CHANGED:
        printf("User has changed the card without unmounting the volume.\n");
        break;
    case FS_OPENFILE:
        printf("File is already open.\n");
        break;
    case FS_DISKFULL:
        printf("Disk is full.\n");
        break;
    case FS_EXIST:
        printf("File already exists.\n");
        break;
    case FS_NOMEM:
        printf("Insufficient memory.\n");
        break;
    case FS_CANTREAD:
        printf("Cannot read disk.\n");
        break;
    case FS_ERROR:
        printf("Hardware error or other unknown error.\n");
        break;
    case FS_USED:
        printf("File/directory is in use.\n");
    }
}



void sh_exit(char *args)
{
    if(shell_data.is_mounted) {
        printf("Must unmount the volume first.\n");
        return;
    }
    shell_data.terminate=1;
}

void sh_mount(char *args)
{
    int error;
    if(shell_data.is_mounted) {
        printf("Already mounted.\n");
        return;
    }
    if(!DRV_OpenDisk()) { sh_error(FS_NOMEM); return ;}

    DRV_GetDiskInfo(&shell_data.dsk);

    //FSFormatVolume(&shell_data.dsk,shell_data.dsk.TotalSectors,4096,(clfs_char *)"DiskLabel");

    FSMemreport();

    error=FSMountVolume(&shell_data.dsk,&shell_data.fs);

    if(error!=FS_OK) {
        sh_error(error);
    } else shell_data.is_mounted=1;

}


void sh_format(char *args)
{
    int error;
    if(shell_data.is_mounted) {
        printf("Disk is mounted, unmount first then format.\n");
        return;
    }
    if(!DRV_OpenDisk()) { sh_error(FS_NOMEM); return ;}

    DRV_GetDiskInfo(&shell_data.dsk);

    error=FSFormatVolume(&shell_data.dsk,shell_data.dsk.TotalSectors,4096,(clfs_char *)"DiskLabel");

    if(error!=FS_OK) {
        sh_error(error);
    }

    DRV_CloseDisk();

    FSMemreport();

}




void sh_unmount(char *args)
{
        int error;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }
    error=FSUnmountVolume(&shell_data.fs,FS_FALSE);

    if(error!=FS_OK) sh_error(error);

    DRV_CloseDisk();

    FSMemreport();

    shell_data.is_mounted=0;


}


void sh_pwd(char *args)
{

    int error;
    clfs_char *path;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    error=FSGetcwd(&path,&shell_data.fs);

    if(error!=FS_OK) {
        sh_error(error);
    }
    else {
    printf("%s\n",path);
    FSFree(path);
    }
   return;
}


void sh_cd(char *args)
{
    int error;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    error=FSChdir((clfs_char *)args,&shell_data.fs);
    if(error!=FS_OK)
    {
        sh_error(error);
    }

    return;
}

void sh_dir(char *args)
{
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    int error;
    printf("Not yet implemented.\n");
    return;
}

void sh_mkdir(char *args)
{
    int error;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    error=FSMkdir((clfs_char *)args,&shell_data.fs);
    if(error!=FS_OK)
    {
        sh_error(error);
    }

    return;
}


void sh_mem(char *args)
{

    FSMemreport();

    return;
}


void sh_check(char *args)
{
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }
    FS_FILE *dir,*entry;
    int error;
    dir=shell_data.fs.CurrentDir;

    FSSeek(dir,0,SEEK_SET);

    do {

    error = FSGetNextDirEntry(dir,&entry);

    if(error!=FS_OK)
    {
        sh_error(error);
        return;
    }


    printf("%c|%c%c%c%c%c %9lld (0x%08llX) ", (FSFileType(entry)==DE_BOOTRECORD)? 'B':((FSFileType(entry)==DE_FILE)? 'F':((FSFileType(entry)==DE_EMPTY)? 'E':((FSFileType(entry)==DE_DELETED)? 'D':'_'))),
           ((FSFileAttributes(entry)&FSATTR_DIR)? 'd':'_'),((FSFileAttributes(entry)&FSATTR_HIDDEN)? 'h':'_'),((FSFileAttributes(entry)&FSATTR_IMMUTABLE)? 'i':'_'),((FSFileAttributes(entry)&FSATTR_CASEINSENS)? 'c':'_'),((FSFileAttributes(entry)&FSATTR_DIRCASEINSENS)? 'C':'_'),FSFileSize(entry),FSFileSize(entry));
    printf("SC=%9lld ",entry->F.StartCluster);
    if(entry->Name==NULL) printf("**(Null-named file)**\n");
    else printf("%s\n",entry->Name);

    FSFreeEntry(entry);

    } while(error==FS_OK);

    return;
}

void sh_ls(char *args)
{
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }
    FS_FILE *dir,*entry;
    int error;
    dir=shell_data.fs.CurrentDir;

    FSSeek(dir,0,SEEK_SET);

    do {

    error = FSGetNextDirEntry(dir,&entry);

    if(error==FS_EOF) break;
    if(error!=FS_OK)
    {
        sh_error(error);
        return;
    }

    if(FSFileType(entry)==DE_FILE) {

    printf("%c%c%c%c%c %12lld ",((FSFileAttributes(entry)&FSATTR_DIR)? 'd':'_'),((FSFileAttributes(entry)&FSATTR_HIDDEN)? 'h':'_'),((FSFileAttributes(entry)&FSATTR_IMMUTABLE)? 'i':'_'),((FSFileAttributes(entry)&FSATTR_CASEINSENS)? 'c':'_'),((FSFileAttributes(entry)&FSATTR_DIRCASEINSENS)? 'C':'_'),FSFileSize(entry));

    if(entry->Name==NULL) printf("**(Null-named file)**\n");
    else printf("%s\n",entry->Name);

    }

    FSFreeEntry(entry);

    } while(error==FS_OK);

    return;
}


void sh_dotest(char *args)
{
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }
int f;
int error;
clfs_char fname[200];
FS_FILE *file;

for(f=0;f<200;++f)
{
    sprintf((char *)fname,"Test File Number %04d",f);
    error=FSCreate(fname,0,&file,&shell_data.fs);
    printf("Creating '%s' = %s\n",fname,FSErrorMsg[-error]);
    if(error==FS_OK)
    error=FSWrite((clfs_byte *)fname,strlen((char *)fname),file);
    if(error!=strlen((char *)fname)) { printf("Can't write to file.\n"); }
    FSClose(file);
}


}

void sh_cat(char *args)
{
int error;
if(!shell_data.is_mounted) {
    printf("No volume was mounted.\n");
    return;
}

struct FS_FILE *file;
error=FSOpen((clfs_char *)args,FSMODE_READ,&file,&shell_data.fs);
if(error!=FS_OK)
{
    sh_error(error);
    return;
}

clfs_64bit f,k;
clfs_char string[33];
for(f=0;f<FSFileSize(file);++f)
{
    error=FSRead((clfs_byte *)string,32,file);

    for(k=0;k<error;++k) putchar(string[k]);
    if(FSEof(file)) break;
}


FSClose(file);
return;
}


void sh_rename(char *args)
{
    int error;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    clfs_char *oldfile,*newfile,*escapedstring,*esc;

    oldfile=(clfs_char *)args;

    int escaped=0;

    escapedstring=esc=(clfs_char *)malloc(strlen(args));
    if(!esc) {
        printf("No memory??\n");
        return;
    }

    while(*oldfile) {
        *esc=*oldfile;
        if( (*oldfile==' ') && (!escaped)) { newfile=esc+1; *esc=0; }
        if( (*oldfile=='\\') && (!escaped)) { escaped=1; --esc; } else escaped=0;
        ++oldfile;
        ++esc;
    }
    *esc=0;

    error=FSRename(escapedstring,newfile,&shell_data.fs);
    if(error!=FS_OK)
    {
        sh_error(error);
    }

    free(escapedstring);
    return;
}


void sh_remove(char *args)
{
    int error;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    clfs_char *oldfile,*newfile,*escapedstring,*esc;

    oldfile=(clfs_char *)args;

    int escaped=0;

    escapedstring=esc=(clfs_char *)malloc(strlen(args)+1);
    if(!esc) {
        printf("No memory??\n");
        return;
    }

    while(*oldfile) {
        *esc=*oldfile;
        if( (*oldfile==' ') && (!escaped)) { newfile=esc+1; *esc=0; }
        if( (*oldfile=='\\') && (!escaped)) { escaped=1; --esc; } else escaped=0;
        ++oldfile;
        ++esc;
    }
    *esc=0;

    error=FSDelete(escapedstring,&shell_data.fs);
    if(error!=FS_OK)
    {
        sh_error(error);
    }

    free(escapedstring);
    return;
}

void sh_rmdir(char *args)
{
    int error;
    if(!shell_data.is_mounted) {
        printf("No volume was mounted.\n");
        return;
    }

    clfs_char *oldfile,*newfile,*escapedstring,*esc;
    int force;
    oldfile=(clfs_char *)args;

    int escaped=0;

    escapedstring=esc=(clfs_char *)malloc(strlen(args)+1);
    if(!esc) {
        printf("No memory??\n");
        return;
    }

    while(*oldfile) {
        *esc=*oldfile;
        if( (*oldfile==' ') && (!escaped)) { newfile=esc+1; *esc=0; }
        if( (*oldfile=='\\') && (!escaped)) { escaped=1; --esc; } else escaped=0;
        ++oldfile;
        ++esc;
    }
    *esc=0;
    if( (escapedstring[0]=='-') && (escapedstring[1]=='f') && (newfile==escapedstring+3)) { force=1; esc=newfile; }
    else { force=0; esc=escapedstring; }
    error=FSRmdir(esc,force,&shell_data.fs);
    if(error!=FS_OK)
    {
        sh_error(error);
    }

    free(escapedstring);
    return;
}
